package core

import (
	"net"
	"os/exec"
	"strconv"
	"syscall"
	"time"

	"github.com/grufwub/go-errors"
)

type envVar struct {
	key   string
	value string
}

func newEnvVar(key, value string) *envVar {
	return &envVar{key, value}
}

func (v *envVar) String() string {
	return v.key + "=" + v.value
}

// getDefaultEnv gets the default (not typed for CGI or SCGI) environment to execute
// under, as a slice of envVar that can easily be used for either
func getDefaultEnv() []*envVar {
	env := make([]*envVar, 0)

	env = append(env, newEnvVar("SERVER_SOFTWARE", "Gophi "+Version))
	env = append(env, newEnvVar("SERVER_PROTOCOL", protocol))
	env = append(env, newEnvVar("SERVER_NAME", Hostname))
	env = append(env, newEnvVar("SERVER_PORT", FwdPort))
	env = append(env, newEnvVar("DOCUMENT_ROOT", Root))

	return env
}

// setupInitialCGIEnv takes a safe PATH, uses other server variables and returns a slice of constant CGI environment variables
func setupInitialCGIEnv(safePath string) []string {
	// Append the default environment
	env := getDefaultEnv()
	env = append(env, newEnvVar("GATEWAY_INTERFACE", "CGI/1.1"))
	env = append(env, newEnvVar("PATH", safePath))

	// Convert to a string slice
	envStrs := make([]string, len(env))
	for i := range env {
		envStrs[i] = env[i].String()
	}

	// Return string slice of environment variables
	return envStrs
}

// setupInitialSCGIEnv uses server variables and returns a slice of constant SCGI environment variables
func setupInitialSCGIEnv() []*envVar {
	// Append the default environment
	env := getDefaultEnv()
	env = append(env, newEnvVar("SCGI", "1"))

	// Return this
	return env
}

// generateCGIEnv takes a Client, and Request object, uses the global constant slice and generates a full set of CGI environment variables
func generateCGIEnv(client *Client, request *Request) []string {
	env := append(cgiEnv, "REMOTE_ADDR="+client.IP())
	env = append(env, "QUERY_STRING="+request.Params())
	env = append(env, "SCRIPT_NAME="+request.Path().Relative())
	env = append(env, "SCRIPT_FILENAME="+request.Path().Absolute())
	env = append(env, "REQUEST_URI="+request.Path().Selector())
	return env
}

// generateSCGIEnv takes a Client, and Request object, uses the global constant slice and generates a full set of SCGI environment variables
func generateSCGIEnv(client *Client, request *Request) []*envVar {
	env := append(scgiEnv, newEnvVar("REMOTE_ADDR", client.IP()))
	env = append(env, newEnvVar("QUERY_STRING", request.Params()))
	env = append(env, newEnvVar("REQUEST_URI", request.Path().Selector()))
	return env
}

// ExecuteCGIScript executes a CGI script, responding with stdout to client
func ExecuteCGIScript(client *Client, request *Request) *errors.Error {
	// Create cmd object
	cmd := exec.Command(request.Path().Absolute())

	// Set new process group id
	cmd.SysProcAttr = &syscall.SysProcAttr{Setpgid: true}

	// Setup cmd environment
	cmd.Env = generateCGIEnv(client, request)
	cmd.Dir = request.Path().Root()

	// Setup cmd out writer
	cmd.Stdout = client.Conn().Writer()

	// Not interested in err
	cmd.Stderr = nil

	// Start executing
	err := cmd.Start()
	if err != nil {
		return errors.WrapError(CGIStartErr, err)
	}

	// Setup goroutine to kill cmd after maxCGIRunTime
	go func() {
		// At least let the script try to finish...
		time.Sleep(maxCGIRunTime)

		// We've already finished
		if cmd.ProcessState != nil {
			return
		}

		// Get process group id
		pgid, err := syscall.Getpgid(cmd.Process.Pid)
		if err != nil {
			SystemLog.Fatalf(pgidNotFoundErrStr)
		}

		// Kill process group!
		err = syscall.Kill(-pgid, syscall.SIGTERM)
		if err != nil {
			SystemLog.Fatalf(pgidStopErrStr, pgid, err.Error())
		}
	}()

	// Wait for command to finish, get exit code
	err = cmd.Wait()
	exitCode := 0
	if err != nil {
		// Error, try to get exit code
		exitError, ok := err.(*exec.ExitError)
		if ok {
			waitStatus := exitError.Sys().(syscall.WaitStatus)
			exitCode = waitStatus.ExitStatus()
		} else {
			exitCode = 1
		}
	} else {
		// No error! Get exit code directly from command process state
		waitStatus := cmd.ProcessState.Sys().(syscall.WaitStatus)
		exitCode = waitStatus.ExitStatus()
	}

	// Non-zero exit code? Return error
	if exitCode != 0 {
		SystemLog.Errorf(cgiExecuteErrStr, request.Path().Absolute(), exitCode)
		return errors.NewError(CGIExitCodeErr)
	}

	// Exit fine!
	return nil
}

// ExecuteSCGI sends an SCGI env request to request SCGI socket
func ExecuteSCGI(client *Client, request *Request) *errors.Error {
	// SCGI socket is stored as request root path
	scgiSocket := request.Path().Root()

	// Connect to SCGI socket, defer close
	conn, err := net.Dial("unix", scgiSocket)
	if err != nil {
		return errors.WrapError(SCGIDialErr, err)
	}
	defer conn.Close()

	// Get SCGI variables + calculate total length
	env := generateSCGIEnv(client, request)
	envBytes := []byte{}
	length := 0
	for _, v := range env {
		// Add the key + zero byte padding, calculate length
		envBytes = append(envBytes, []byte(v.key+"\x00")...)
		length += len(v.key) + 1
		// Add the value + zero byte padding, calculate length
		envBytes = append(envBytes, []byte(v.value+"\x00")...)
		length += len(v.value) + 1
	}

	// Prepend envBytes with expected length, then append ','
	envBytes = append([]byte(strconv.Itoa(length)+":"), envBytes...)
	envBytes = append(envBytes, ',')

	// Send SCGI environment (we don't bother wrapping in buffered
	// writer the request env will be of limited size, and we send it
	// all at once). Set deadline for this single write
	conn.SetWriteDeadline(time.Now().Add(scgiWriteDeadline))
	_, err = conn.Write(envBytes)
	if err != nil {
		return errors.WrapError(SCGIWriteErr, err)
	}

	// Wrap conn in BufferedReader, defer putting back
	br := scgiBufferedReaderPool.Get(conn)
	defer scgiBufferedReaderPool.Put(br)

	// Set conn read deadline
	conn.SetReadDeadline(time.Now().Add(scgiReadDeadline))

	// Connect client directly to socket output
	return client.Conn().ReadFrom(br)
}
