package core

import (
	"flag"
	"fmt"
	"log"
	"os"
	"os/signal"
	"path"
	"strconv"
	"strings"
	"syscall"
	"time"

	"github.com/grufwub/go-bufpools"
	"github.com/grufwub/go-errors"
	"github.com/grufwub/go-filecache"
	"github.com/grufwub/go-logger"
)

// ParseFlagsAndSetup parses necessary core server flags, and sets up the core ready for Start() to be called
func ParseFlagsAndSetup(proto string) {
	// Setup numerous temporary flag variables, and store the rest
	// directly in their final operating location. Strings are stored
	// in `string_constants.go` to allow for later localization
	sysLog := flag.String(sysLogFlagStr, "stdout", sysLogDescStr)
	accLog := flag.String(accLogFlagStr, "stdout", accLogDescStr)
	flag.StringVar(&Root, rootFlagStr, "/var/gopher", rootDescStr)
	flag.StringVar(&Bind, bindFlagStr, "", bindDescStr)
	flag.StringVar(&Hostname, hostnameFlagStr, "localhost", hostnameDescStr)
	port := flag.Uint(portFlagStr, 70, portDescStr)
	fwdPort := flag.Uint(fwdPortFlagStr, 0, fwdPortDescStr)
	flag.DurationVar(&connReadDeadline, connReadTimeoutFlagStr, time.Duration(time.Second*3), connReadTimeoutDescStr)
	flag.DurationVar(&connWriteDeadline, connWriteTimeoutFlagStr, time.Duration(time.Second*5), connWriteTimeoutDescStr)
	cReadBuf := flag.Uint(connReadBufFlagStr, 1024, connReadBufDescStr)
	cWriteBuf := flag.Uint(connWriteBufFlagStr, 1024, connWriteBufDescStr)
	cReadMax := flag.Uint(connReadMaxFlagStr, 4096, connReadMaxDescStr)
	fReadBuf := flag.Uint(fileReadBufFlagStr, 1024, fileReadBufDescStr)
	flag.DurationVar(&monitorSleepTime, monitorSleepTimeFlagStr, time.Duration(time.Second*1), monitorSleepTimeDescStr)
	cacheMax := flag.Float64(cacheFileMaxFlagStr, 1.0, cacheFileMaxDescStr)
	cacheSize := flag.Uint(cacheSizeFlagStr, 100, cacheSizeDescStr)
	cacheAgeMax := flag.Duration(cacheAgeMaxFlagStr, time.Minute*5, cacheAgeMaxDescStr)
	restrictedPathsList := flag.String(restrictPathsFlagStr, "", restrictPathsDescStr)
	hiddenPathsList := flag.String(hiddenPathsFlagStr, "", hiddenPathsDescStr)
	remapRequestsList := flag.String(remapRequestsFlagStr, "", remapRequestsDescStr)
	cgiDir := flag.String(cgiDirFlagStr, "", cgiDirDescStr)
	scgiPathsList := flag.String(scgiPathsFlagStr, "", scgiPathsDescStr)
	scgiReadBuf := flag.Uint(scgiReadBufFlagStr, 1024, scgiReadBufDescStr)
	flag.DurationVar(&scgiReadDeadline, scgiReadTimeoutFlagStr, time.Duration(time.Second*1), scgiReadTimeoutDescStr)
	flag.DurationVar(&scgiWriteDeadline, scgiWriteTimeoutFlagStr, time.Duration(time.Second*3), scgiWriteTimeoutDescStr)
	flag.DurationVar(&maxCGIRunTime, maxCGITimeFlagStr, time.Duration(time.Second*3), maxCGITimeDescStr)
	safePath := flag.String(safePathFlagStr, "/bin:/usr/bin", safePathDescStr)
	flag.StringVar(&userDir, userDirFlagStr, "", userDirDescStr)
	printVersion := flag.Bool(versionFlagStr, false, versionDescStr)

	// Parse flags! (including any set by outer calling function)
	flag.Parse()

	// If version print requested, do so!
	if *printVersion {
		fmt.Println("Gophi " + Version)
		os.Exit(0)
	}

	// Setup loggers
	SystemLog = setupLogger(*sysLog)
	if *sysLog == *accLog {
		AccessLog = SystemLog
	} else {
		AccessLog = setupLogger(*accLog)
	}

	// Check valid values for BindAddr and Hostname
	if Hostname == "" {
		if Bind == "" {
			SystemLog.Fatal(hostnameBindEmptyStr)
		}
		Hostname = Bind
	}

	// Change to server directory
	if osErr := os.Chdir(Root); osErr != nil {
		SystemLog.Fatalf(chDirErrStr, osErr)
	}
	SystemLog.Infof(chDirStr, Root)

	// Set port info
	if *fwdPort == 0 {
		fwdPort = port
	}
	Port = strconv.Itoa(int(*port))
	FwdPort = strconv.Itoa(int(*fwdPort))

	// Set protocol string (only really used by CGI and one call to SystemLog)
	protocol = proto

	// Setup listener
	var err *errors.Error
	serverListener, err = newListener(Bind, Port)
	if err != nil {
		SystemLog.Fatalf(listenerBeginFailStr, protocol, Hostname, FwdPort, Bind, Port, err.Error())
	}

	// Setup the sync pools
	connBufferedReaderPool = bufpools.NewBufferedReaderPool(int(*cReadBuf))
	connBufferedWriterPool = bufpools.NewBufferedWriterPool(int(*cWriteBuf))
	fileBufferedReaderPool = bufpools.NewBufferedReaderPool(int(*fReadBuf))
	fileBufferPool = bufpools.NewBufferPool(int(*fReadBuf))

	// Conn read max
	connReadMax = int(*cReadMax)

	// FileSystemObject (and related) setup
	fileSizeMax = int64(1048576.0 * *cacheMax) // gets megabytes value in bytes
	FileCache = filecache.NewFileCache(int(*cacheSize), *cacheAgeMax)

	// If no restricted paths provided, set to the disabled function. Else, compile and enable
	if *restrictedPathsList == "" {
		SystemLog.Info(pathRestrictionsDisabledStr)
		IsRestrictedPath = isRestrictedPathDisabled
	} else {
		SystemLog.Info(pathRestrictionsEnabledStr)
		restrictedPaths = compileRestrictedPathsRegex(*restrictedPathsList)
		IsRestrictedPath = isRestrictedPathEnabled
	}

	// If no hidden paths provided, set to the disabled function. Else, compile and enable
	if *hiddenPathsList == "" {
		SystemLog.Info(pathHidingDisableStr)
		IsHiddenPath = isHiddenPathDisabled
	} else {
		SystemLog.Info(pathHidingEnabledStr)
		hiddenPaths = compileHiddenPathsRegex(*hiddenPathsList)
		IsHiddenPath = isHiddenPathEnabled
	}

	// If no remapped paths provided, set to the disabled function. Else, compile and enable
	if *remapRequestsList == "" {
		SystemLog.Info(requestRemapDisabledStr)
		RemapRequest = remapRequestDisabled
	} else {
		SystemLog.Info(requestRemapEnabledStr)
		requestRemaps = compileRequestRemapRegex(*remapRequestsList)
		RemapRequest = remapRequestEnabled
	}

	// If no CGI dir supplied, set to disabled function. Else, compile and enable
	if *cgiDir == "" {
		SystemLog.Info(cgiSupportDisabledStr)
		WithinCGIDir = withinCGIDirDisabled
	} else {
		SystemLog.Info(cgiSupportEnabledStr)
		cgiDirRegex = compileCGIRegex(*cgiDir)
		cgiEnv = setupInitialCGIEnv(*safePath)
		WithinCGIDir = withinCGIDirEnabled
	}

	// If no SCGI paths provided, set to disabled function. Else, compile and enable
	if *scgiPathsList == "" {
		SystemLog.Info(scgiSupportDisabledStr)
		MapSCGIRequest = mapSCGIRequestDisabled
	} else {
		SystemLog.Info(scgiSupportEnabledStr)
		scgiMaps = compileSCGIMapRegex(*scgiPathsList)
		scgiEnv = setupInitialSCGIEnv()
		MapSCGIRequest = mapSCGIRequestEnabled
		scgiBufferedReaderPool = bufpools.NewBufferedReaderPool(int(*scgiReadBuf))
	}

	// If no user dir supplied, set to disabled function. Else, set user dir and enable
	if userDir == "" {
		SystemLog.Info(userDirDisabledStr)
		getRequestPath = getRequestPathUserDirDisabled
	} else {
		SystemLog.Info(userDirEnabledStr)
		getRequestPath = getRequestPathUserDirEnabled

		// Clean the user dir to be safe
		userDir = path.Clean(userDir)
		if strings.HasPrefix(userDir, "..") {
			SystemLog.Fatalf(userDirBackTraverseErrStr, userDir)
		} else {
			SystemLog.Infof(userDirStr, userDir)
		}
	}

	// Setup signal channel
	sigChannel = make(chan os.Signal)
	signal.Notify(sigChannel, syscall.SIGINT, syscall.SIGTERM, syscall.SIGKILL)
}

func setupLogger(output string) *logger.SLogger {
	switch output {
	case "stdout":
		return logger.NewSLogger(os.Stdout, true)
	case "stderr":
		return logger.NewSLogger(os.Stderr, true)
	case "null":
		return logger.NewSLogger(&logger.NilWriter{}, true)
	default:
		file, err := os.OpenFile(output, os.O_CREATE|os.O_APPEND|os.O_WRONLY, 0600)
		if err != nil {
			log.Fatalf(logOutputErrStr, output, err.Error())
		}
		return logger.NewSLogger(file, true)
	}
}
