package core

import (
	"os"
	"regexp"
	"time"

	"github.com/grufwub/go-bufpools"
	"github.com/grufwub/go-errors"
	"github.com/grufwub/go-filecache"
	"github.com/grufwub/go-logger"
)

const (
	// Version holds the current version string
	Version = "v2.5.2-beta"
)

var (
	// Root stores the server's root directory
	Root string

	// Bind stores the server's bound IP
	Bind string

	// Hostname stores the host's outward hostname
	Hostname string

	// Port stores the internal port the host is binded to
	Port string

	// FwdPort stores the host's outward port number
	FwdPort string

	// AccessLog is the global Access SLogger
	AccessLog *logger.SLogger

	// SystemLog is the global System SLogger
	SystemLog *logger.SLogger

	// FileCache is the global FileCache object
	FileCache *filecache.FileCache

	// File system related globals
	monitorSleepTime time.Duration
	fileSizeMax      int64
	userDir          string

	// Global listener
	serverListener *listener

	// Client connection related globals
	connReadDeadline  time.Duration
	connWriteDeadline time.Duration
	connReadMax       int

	// Server protocol
	protocol string

	// CGI related globals
	cgiEnv        []string
	maxCGIRunTime time.Duration

	// SCGI related globals
	scgiEnv           []*envVar
	scgiReadDeadline  time.Duration
	scgiWriteDeadline time.Duration

	// Global OS signal channel
	sigChannel chan os.Signal

	// Buffer pools
	connBufferedReaderPool *bufpools.BufferedReaderPool
	connBufferedWriterPool *bufpools.BufferedWriterPool
	scgiBufferedReaderPool *bufpools.BufferedReaderPool
	fileBufferedReaderPool *bufpools.BufferedReaderPool
	fileBufferPool         *bufpools.BufferPool

	// Compiled regex globals
	cgiDirRegex     *regexp.Regexp
	restrictedPaths []*regexp.Regexp
	hiddenPaths     []*regexp.Regexp
	requestRemaps   []*RequestRemap
	scgiMaps        []*RequestRemap

	// WithinCGIDir returns whether a path is within the server's specified CGI scripts directory
	WithinCGIDir func(*Path) bool

	// IsRestrictedPath is the global function to check against restricted paths
	IsRestrictedPath func(*Path) bool

	// IsHiddenPath is the global function to check against hidden paths
	IsHiddenPath func(*Path) bool

	// RemapRequest is the global function to remap a request
	RemapRequest func(*Request) bool

	// MapSCGIRequest is the global function to map a request to SCGI
	MapSCGIRequest func(*Request) bool
)

// Start begins operation of the server
func Start(serve func(*Client)) {
	// Start the FileSystemObject cache freshness monitor
	SystemLog.Infof(cacheMonitorStartStr, monitorSleepTime)
	go FileCache.StartMonitor(monitorSleepTime)

	// Start the listener
	SystemLog.Infof(listeningOnStr, protocol, Hostname, FwdPort, Bind, Port)
	go func() {
		for {
			client, err := serverListener.Accept()
			if err != nil {
				SystemLog.Errorf(err.Error())
			}

			// Serve client then close in separate goroutine
			go func() {
				serve(client)
				client.Conn().Close()
			}()
		}
	}()

	// Listen for OS signals and terminate if necessary
	sig := <-sigChannel
	SystemLog.Infof(signalReceivedStr, sig)
	os.Exit(0)
}

// HandleClient handles a Client, attempting to serve their request from the filesystem whether a regular file, gophermap, dir listing or CGI script
func HandleClient(client *Client, request *Request, newFileContents func(*Path) FileContent, handleDirectory func(*Client, *os.File, *Path) *errors.Error) *errors.Error {
	// If restricted, return error
	if IsRestrictedPath(request.Path()) {
		return errors.NewError(RestrictedPathErr)
	}

	// If SCGI path map to socket, else try remap request
	if ok := MapSCGIRequest(request); ok {
		client.LogInfo(scgiRequestMappedStr, request.Path().Selector(), request.Params())
		return ExecuteSCGI(client, request)
	} else if ok := RemapRequest(request); ok {
		client.LogInfo(requestRemappedStr, request.Path().Selector(), request.Params())
	}

	// First check for file on disk
	file, err := OpenFile(request.Path())
	if err != nil {
		// Get read-lock, defer unlock
		FileCache.RLock()
		defer FileCache.RUnlock()

		// Don't throw in the towel yet! Check for generated file in cache
		cached, ok := FileCache.Get(request.Path().Absolute())
		if !ok {
			return err
		}

		// We got a generated file! Get content
		content := cached.Content().(FileContent)

		// Send content to client
		return content.WriteToClient(client, request.Path())
	}
	defer file.Close()

	// Get stat
	stat, goErr := file.Stat()
	if goErr != nil {
		return errors.WrapError(FileStatErr, goErr)
	}

	switch {
	// Directory
	case stat.Mode()&os.ModeDir != 0:
		// Don't support CGI script dir enumeration
		if WithinCGIDir(request.Path()) {
			return errors.NewError(RestrictedPathErr)
		}

		// Else enumerate dir
		return handleDirectory(client, file, request.Path())

	// Regular file
	case stat.Mode()&os.ModeType == 0:
		// Execute script if within CGI dir
		if WithinCGIDir(request.Path()) {
			return ExecuteCGIScript(client, request)
		}

		// Else just fetch
		return FetchFile(client, file, stat, request.Path(), newFileContents)

	// Unsupported type
	default:
		return errors.NewError(FileTypeErr)
	}
}

// FetchFile attempts to fetch a file from the cache, using the supplied file stat, Path and serving client. Returns Error status
func FetchFile(client *Client, file *os.File, stat os.FileInfo, p *Path, newFileContent func(*Path) FileContent) *errors.Error {
	// If file too big, write direct to client
	if stat.Size() > fileSizeMax {
		return client.Conn().ReadFrom(file)
	}

	// Get cache read lock
	FileCache.RLock()

	// Define cache here
	var content FileContent

	// Now check for file in cache
	cached, ok := FileCache.Get(p.Absolute())
	if !ok {
		// Create new file contents with supplied function
		content = newFileContent(p)

		// Cache the file contents
		err := content.Load(p, file)
		if err != nil {
			// Unlock, return error
			FileCache.RUnlock()
			return err
		}

		// Wrap contents in file, set fresh
		cached = filecache.NewFile(p.Absolute(), true, content)
		cached.UpdateLastRefresh()

		// Try upgrade our lock, else error out (have to remember to unlock!!)
		if !FileCache.UpgradeLock() {
			FileCache.Unlock()
			return errors.NewError(MutexUpgradeErr)
		}

		// Put file in cache
		FileCache.Put(cached)

		// Try downgrade our lock, else error out (have to remember to runlock!!)
		if !FileCache.DowngradeLock() {
			FileCache.RUnlock()
			return errors.NewError(MutexDowngradeErr)
		}

		// Get file read lock
		cached.RLock()
	} else {
		// Get file read lock
		cached.RLock()

		// Get contents from file
		content = cached.Content().(FileContent)

		// Check for file freshness
		if !cached.IsFresh() {
			// Try upgrade file lock, else error out (have to remember to unlock!!)
			if !FileCache.UpgradeLock() {
				FileCache.Unlock()
				return errors.NewError(MutexUpgradeErr)
			}

			// Refresh file contents
			err := content.Load(p, file)
			if err != nil {
				// Unlock file, return error
				FileCache.Unlock()
				return err
			}

			// Set file as fresh
			cached.UpdateLastRefresh()

			// Try downgrade file lock, else error out (have to remember to runlock!!)
			if !FileCache.DowngradeLock() {
				FileCache.RUnlock()
				return errors.NewError(MutexDowngradeErr)
			}
		}
	}

	// Defer file + cache read unlock
	defer func() {
		cached.RUnlock()
		FileCache.RUnlock()
	}()

	// Write to client
	return content.WriteToClient(client, p)
}
