package gopher

import (
	"gophi/core"
	"os"
	"strings"

	"github.com/grufwub/go-errors"
)

var (
	// Gophermap formatting globals
	subgophermapSizeMax int64
	pageWidth           int
	footer              []byte
)

// serve is the global gopher server's serve function
func serve(client *core.Client) {
	// Receive line from client
	received, err := client.Conn().ReadLine()
	if err != nil {
		client.LogError(clientReadFailStr)
		handleError(client, err)
		return
	}

	// Convert to string + remove leading '/'
	line := strings.TrimPrefix(string(received), "/")

	// If prefixed by 'URL:' send a redirect
	lenBefore := len(line)
	line = strings.TrimPrefix(line, "URL:")
	if len(line) < lenBefore {
		client.Conn().Write(generateHTMLRedirect(line))
		client.LogInfo(clientRedirectFmtStr, line)
		return
	}

	// Parse new request
	request, err := core.ParseURLEncodedRequest(line)
	if err != nil {
		client.LogError(clientRequestParseFailStr)
		handleError(client, err)
		return
	}

	// Handle the request!
	err = core.HandleClient(
		// Current client
		client,

		// Current request
		request,

		// New file contents function
		newFileContent,

		// Handle directory function
		func(client *core.Client, file *os.File, p *core.Path) *errors.Error {
			// First check for gophermap, create gophermap Path object
			gophermap := p.JoinPath("gophermap")

			// If gophermap exists, we fetch this
			file2, err := core.OpenFile(gophermap)
			if err == nil {
				stat, osErr := file2.Stat()
				if osErr == nil {
					// Fetch gophermap and defer close
					defer file2.Close()
					return core.FetchFile(client, file2, stat, gophermap, newFileContent)
				}

				// Else, just close file2
				file2.Close()
			}

			// Slice to write
			dirContents := make([]byte, 0)

			// Add directory heading + empty line
			dirContents = append(dirContents, buildLine(typeInfo, "[ "+core.Hostname+p.Selector()+" ]", "TITLE", nullHost, nullPort)...)
			dirContents = append(dirContents, buildInfoLine("")...)

			// Scan directory and build lines
			err = core.ScanDirectory(
				// Directory file
				file,

				// Directory path
				p,

				// Iter function
				func(file os.FileInfo, fp *core.Path) {
					// Append new formatted file listing (if correct type)
					dirContents = appendFileListing(dirContents, file, fp)
				},
			)
			if err != nil {
				return err
			}

			// Add footer, write contents
			dirContents = append(dirContents, footer...)
			return client.Conn().Write(dirContents)
		},
	)

	// Final error handling
	if err != nil {
		handleError(client, err)
		client.LogError(clientServeFailStr, request.String())
	} else {
		client.LogInfo(clientServedStr, request.String())
	}
}

// handleError determines whether to send an error response to the client, and logs to system
func handleError(client *core.Client, err *errors.Error) {
	response, ok := generateErrorResponse(err)
	if ok {
		client.Conn().Write(response)
	}
	core.SystemLog.Errorf(err.Error())
}

// newFileContents returns a new FileContents object
func newFileContent(p *core.Path) core.FileContent {
	if isGophermap(p) {
		return &gophermapContent{}
	}
	return &core.RegularFileContent{}
}
